# code support tractometric statistical analysis for dipy

import warnings
w_string = "The `dipy.stats` module is still under heavy development "
w_string += "and functionality, as well as the API is likely to change "
w_string += "in future versions of the software"
warnings.warn(w_string)
